#!/usr/bin/perl

if($ARGV[0] eq "")
{
	print "USAGE:   create_net.pl [options]\n";
	print "\n";	
	print "OPTIONS: -o file        Network for the SVite tool\n";
	print "         -p file        Phoneme list (if a loop without any LM is created)\n";	
	print "         -a file        ARPA LM file\n";
	print "         -0                use only phoneme list from the ARPA LM file\n";
	print "         -1                use only 1-grams from the ARPA LM file\n";
	print "         -n start end   Start and end symbols (are discarded)\n";
	print "         -l file        Lexicon\n";
	exit(1);
}

$output_file = "";
$arpa_file = "";
$phoneme_file = "";
$lexicon_file = "";
$only_phonemes = 0;
$only_1grams = 0;
$start_symbol = "!ENTER";
$end_symbol = "!EXIT";

while($ARGV[0] ne "")
{	
	if($ARGV[0] eq "-o")
	{
		$output_file = $ARGV[1];
		shift; 
	} 
	if($ARGV[0] eq "-a")
	{
		$arpa_file = $ARGV[1];
		shift; 
	}
	if($ARGV[0] eq "-p")
	{
		$phoneme_file = $ARGV[1];
		shift; 
	}
	if($ARGV[0] eq "-l")
	{
		$lexicon_file = $ARGV[1];
		shift; 
	}	
	if($ARGV[0] eq "-0")
	{
		$only_phonemes = 1;
	}
	if($ARGV[0] eq "-1")
	{
		$only_1grams = 1;
	}
	if($ARGV[0] eq "-n")
	{
		$start_symbol = $ARGV[1];
		$end_symbol = $ARGV[2];
		shift; shift;
	}
	shift;
}

if($arpa_file ne "" && $phoneme_file ne "")
{
	print "ERROR: You can not use both ARPA language model and pure phoneme list.\n" ;
	exit(1);
}


# -- load the ARPA LM --------------------------------------------------------------------------------

if($arpa_file ne "")
{
	open LM, $arpa_file or die "Can not open the ARPA LM file: $arpa_file";
	
	$data = <LM>;
	chomp $data;
	if(! $data =~ /\\data\\/)
	{
		print "ERROR: Invalid ARPA LM file: $arpa_file\n";
		exit(1);
	}
	
	# get ngram order (n) and numbers of 1-grams, 2-grams, ..., n-grams
	do
	{
		$l = <LM>;
		chomp $l;
		if($l ne "")
		{
			$l =~ /ngram (.+)=(.+)/;
			$n=$1;
			$ngrams[$n]=$2;
		}
	}while($l ne "");
	
	# load the LM
	%labels = ();
	while(!eof(LM))
	{
		do
		{
			$l = <LM>;
			chomp $l;
		} while($l eq "");
		
		last if ($l eq "\\end\\");
		
		$ok = ($l =~ /\\(.+)-grams:/);
		if (!ok)
		{
			print "ERROR: Can not find '\\X-grams:' tag in the ARPA LM file.\n";
			exit(1);
		}
		$order=$1;
		for($i = 0; $i < $ngrams[$order]; $i++)
		{
			$l = <LM>;
			chomp $l;
			if($l eq "")
			{
				print "ERROR: There is less $order-gram than number specified in the header in the APRA LM file.\n";
				exit(1);
			}
			@columns = split /\s+/, $l;
			$label="";

			$has_backoff = ($columns[@columns - 1] =~ /^[0-9.+-]+$/);
			$to = ($has_backoff ? @columns - 1 : @columns);	
			
			for($j = 1; $j < $to; $j++)
			{
				if($label eq "")
				{
					$label = "$columns[$j]";
				}
				else
				{
					$label = "$label $columns[$j]";
				}
			}
			
			#print "$label\n";
			$prob{$label} = $columns[0] / log(10);
			$backoff{$label} = $columns[@columns - 1] / log(10) if($has_backoff);
			$labels{$label} = 1 if($order == 1);
		}
	}
	close LM;
}

# -- load pure phoneme list -------------------------------------------------------------------------------------

if($phoneme_file ne "")
{
	open L, $phoneme_file or die "Can not open the phoneme list";
	%labels = ();
	while($l = <L>)
	{
		chomp $l;
		$labels{$l} = 1;
	}
	close L;
}

# -- load or create lexicon -------------------------------------------------------------------------------------

if($lexicon_file ne "")
{
	open L, $lexicon_file or die "Can not open the lexicon file";
	%lexicon = ();
	while($l = <L>)
	{
		chomp $l;
		($word, $trans) = split /\s+/, $l, 2;
		$lexicon{$word} = $trans;
	}
	close L;
}
else
{
	foreach $lab (keys %labels)
	{
		if($lab ne $start_symbol && $lab ne $end_symbol)
		{
			$lexicon{$lab} = $lab;
			#print "$lexicon{$lab} $lab\n";
		}
	}
}

# -- functions for costruction of the LM network ----------------------------------------------------

%units = ();
$freeID = 0;

sub createRecord
{	
	my $id = $_[0];
	my $type = $_[1];
	my $wrd_mdl = $_[2];
	my $links = $_[3];
	my $probs = $_[4];
	my $rec = {};
	$rec->{id} = $id;	
	$rec->{type} = $type;
	$rec->{wrd_mdl} = $wrd_mdl;
	$rec->{links} = $links;
	$rec->{probs} = $probs;
	return $rec
}

sub addLink
{
	my $start = $_[0];
	my $end = $_[1];
	my $prob = $_[2];
	if(! defined($units{$start}))
	{
		print "ERROR: addLink - The start node $start does not exist.\n";
		exit(1);
	}
	if(! defined($units{$end}))
	{
		print "ERROR: addLink - The end node $end does not exist.\n";
		exit(1);
	}	
	my $rec1 = $units{$start};
	
	my $links = $rec1->{links};
	my $probs = $rec1->{probs};
	if($links eq "")
	{
		$rec1->{links} = "$end";
		$rec1->{probs} = "$prob";
	}
	else
	{
		$rec1->{links} = "$links $end";
		$rec1->{probs} = "$probs $prob";	
	}
}

sub addWord
{
	my $name = $_[0];
	
	if(!defined($lexicon{$name}))
	{
		print "ERROR: Can not find pronunciation of the word '$name'.\n";
		exit(1);
	}	
	$pron = $lexicon{$name};
	@labels = split /\s+/, $pron;
	$i = 0;
	$first = 0;
	$last = 0;
	$pid = -1;
	
	foreach $lab (@labels)
	{
		$id = addModel($lab);
		$first = $id if($i == 0);
		$last = $id if($i == @labels - 1);
		addLink($pid, $id, 0) if($pid != -1);
		$i++;
		$pid = $id;
	}
	my $rec = createRecord($freeID, "W", $name, "", "");
	$units{$freeID} = $rec;
	$freeID++;
	addLink($last, $freeID - 1, 0);
	$last = $freeID - 1;
	return ($first, $last);
}

sub addModel
{
	my $name = $_[0];	
	my $rec = createRecord($freeID, "M", $name, "", "");
	$units{$freeID} = $rec;
	$freeID++;
}

sub addNode
{
	my $name = $_[0];
	my $rec = createRecord($freeID, "N", "-", "", "");
	$units{$freeID} = $rec;
	$freeID++;
	return $freeID - 1;
}

sub writeNet
{
	# calculate the number of arcs 
	$narcs = 0;
	foreach $u (keys %units)
	{
		@links1 = split /\s+/, $units{$u}->{links};
		$narcs += @links1;
	}

	# dump the net
	my $file = $_[0];
	open OUT, ">$file" or die "Can not write the network to the file: $file";
	$nnodes = (keys %units) + 0;
	#print "NumberOfNodes=$nnodes NumberOfArcs=$narcs\n";
	print OUT "NumberOfNodes=$nnodes NumberOfArcs=$narcs\n";
	foreach $u (sort {$a <=> $b} keys %units)
	{
		my $rec = $units{$u};
		my $id = $rec->{id};
		my $type = $rec->{type};
		my $wrd_mdl = $rec->{wrd_mdl};
		my $links = $rec->{links};
		my $probs = $rec->{probs};
		@links1 = split /\s+/, $links;
		@probs1 = split /\s+/, $probs;
		$nlinks = @links1;
		$links2 = "";
		for($i = 0; $i < @links1; $i++)
		{
			if($links2 eq "")
			{
				$links2 = $links1[$i];
			}
			else
			{
				$links2 = "$links2 $links1[$i]";
			}
			if($probs1[$i] != 0)
			{
				$links2 = "$links2 {$probs1[$i]}";
			}
		}
		
		#printf "$id\t$type\t%8s\t\t\t\t$nlinks\t$links2\n", $wrd_mdl;
		printf OUT "$id\t$type\t%8s\t\t\t\t$nlinks\t$links2\n", $wrd_mdl;
	}
	close OUT;
}

# -- construct the net ------------------------------------------------------------------------

delete $labels{$start_symbol};
delete $labels{$end_symbol};
$nphn=(keys %labels) + 0;

# create pure phoneme loop
if($phoneme_file ne "" || $only_phonemes == 1)
{

	$start_node = addNode("-");
	$end_node = addNode("-");
	$help_node = addNode("-");

	foreach $w (keys %labels)
	{
		($id_start, $id_end) = addWord($w);
		addLink($id_end, $help_node, 0);
		addLink($help_node, $id_start, 0);
		addLink($start_node, $id_start, 0);
	}

	addLink($help_node, $end_node, 0);
	writeNet($output_file);
	exit(0);
}

# create a net with 1grams
if($arpa_file ne "" && $only_1grams == 1)
{
	$start_node = addNode("-");
	$end_node = addNode("-");
	$help_node = addNode("-");

	foreach $w (keys %labels)
	{
		($id_start, $id_end) = addWord($w);

		addLink($id_end, $help_node, 0);
		$p = $prob{$w};
		addLink($help_node, $id_start, $p);
		addLink($start_node, $id_start, $p);
	}

	addLink($help_node, $end_node, 0);
	writeNet($output_file);
}
else    # net with bigrams
{
	$start_node = addNode("-");
	$end_node = addNode("-");	

	%model_starts = ();
	%model_ends = ();
	%backoff_nodes = ();
	foreach $w (keys %labels)
	{
		($id_start, $id_end) = addWord($w);
		$model_starts{$w} = $id_start;
		$model_ends{$w} = $id_end;
		
		addLink($start_node, $id_start, 0);
		addLink($id_end, $end_node, 0);
	}
	
	foreach $w1 (keys %labels)
	{
		foreach $w2 (keys %labels)
		{
			if(defined($prob{"$w1 $w2"}))   # use bibram
			{
				$start = $model_ends{$w1};
				$end = $model_starts{$w2};
				$p = $prob{"$w1 $w2"};
				#print "$w1 $w2 $start $end\n";
				addLink($start, $end, $p); #$p
			}
			#else                            # use backoff node
			#{
			#}
		}
	}
	
	if($output_file ne "")
	{
		writeNet($output_file);
	}
}


